from os import environ
from pathlib import Path
from typing import Optional, Dict, List

from phidata import constants
from phidata.product import DataProduct
from phidata.workflow import Workflow
from phidata.task import Task
from phidata.workspace import WorkspaceConfig
from phidata.types.context import PathContext, RunContext

from phiterm.types.run_status import RunStatus
from phiterm.utils.cli_console import (
    print_info,
    print_heading,
    print_dict,
    print_info,
    print_subheading,
)
from phiterm.utils.log import logger
from phiterm.workspace.phi_ws_data import PhiWsData


def set_local_env(
    ws_config: WorkspaceConfig,
    path_context: PathContext,
) -> None:
    """VERY IMPORTANT TO RUN BEFORE RUNNING ANY LOCAL WORKFLOWS
    This function updates the local environment with Paths and Configuration
    Main uses:
        1. Path env variables used by DataAssets for building file_paths
        2. Cloud config variables used by AwsAssets like aws_region and profile
    """
    # 1. Path env variables used by DataAssets for building file_paths
    if path_context.scripts_dir is not None:
        environ[constants.SCRIPTS_DIR_ENV_VAR] = str(path_context.scripts_dir)
    if path_context.storage_dir is not None:
        environ[constants.STORAGE_DIR_ENV_VAR] = str(path_context.storage_dir)
    if path_context.meta_dir is not None:
        environ[constants.META_DIR_ENV_VAR] = str(path_context.meta_dir)
    if path_context.products_dir is not None:
        environ[constants.PRODUCTS_DIR_ENV_VAR] = str(path_context.products_dir)
    if path_context.notebooks_dir is not None:
        environ[constants.NOTEBOOKS_DIR_ENV_VAR] = str(path_context.notebooks_dir)
    if path_context.workspace_config_dir is not None:
        environ[constants.WORKSPACE_CONFIG_DIR_ENV_VAR] = str(
            path_context.workspace_config_dir
        )
    # 2. Cloud config variables used by AwsAssets like aws_region and profile
    if ws_config.aws_region is not None:
        environ[constants.AWS_REGION_ENV_VAR] = ws_config.aws_region
    if ws_config.aws_profile is not None:
        environ[constants.AWS_PROFILE_ENV_VAR] = ws_config.aws_profile
    if ws_config.aws_config_file is not None:
        environ[constants.AWS_CONFIG_FILE_ENV_VAR] = ws_config.aws_config_file
    if ws_config.aws_shared_credentials_file is not None:
        environ[
            constants.AWS_SHARED_CREDENTIALS_FILE_ENV_VAR
        ] = ws_config.aws_shared_credentials_file
    if ws_config.local_env:
        environ.update(ws_config.local_env)


def run_workflows_local(
    workflow_file: str,
    run_context: RunContext,
    ws_data: PhiWsData,
    workflows: Dict[str, Workflow],
    workflow_name: Optional[str] = None,
    tasks: Optional[Dict[str, Task]] = None,
    task_name: Optional[str] = None,
) -> bool:
    """
    Runs a Workflow in the local environment

    Args:
        workflow_file: Path of the workflow file relative to the products_dir.
            This is used to build the path_context
        run_context: RunContext
        ws_data: PhiWsData
        workflows: Dict[str, Workflow] to run
        workflow_name: Name of workflow to run
        tasks: Dict[str, Task]
        task_name:Name of task to run

    Returns:
        run_status (bool): True is the Workflow ran successfully
    """
    from phidata.utils.cli_console import print_run_status_table

    # Validate
    if ws_data.ws_config is None:
        logger.error("WorkspaceConfig invalid")
        return False
    if ws_data.ws_root_path is None:
        logger.error("Workspace directory invalid")
        return False

    # Step 1: Build the PathContext for the workflows.
    ws_config = ws_data.ws_config
    ws_root_path = ws_data.ws_root_path
    scripts_dir_path = ws_root_path.joinpath(ws_config.scripts_dir).resolve()
    storage_dir_path = ws_root_path.joinpath(ws_config.storage_dir).resolve()
    meta_dir_path = ws_root_path.joinpath(ws_config.meta_dir).resolve()
    products_dir_path = ws_root_path.joinpath(ws_config.products_dir).resolve()
    notebooks_dir_path = ws_root_path.joinpath(ws_config.notebooks_dir).resolve()
    workspace_config_dir_path = ws_root_path.joinpath(
        ws_config.workspace_config_dir
    ).resolve()
    workflow_file_path = Path(products_dir_path).joinpath(workflow_file)
    path_context: PathContext = PathContext(
        scripts_dir=scripts_dir_path,
        storage_dir=storage_dir_path,
        meta_dir=meta_dir_path,
        products_dir=products_dir_path,
        notebooks_dir=notebooks_dir_path,
        workspace_config_dir=workspace_config_dir_path,
        workflow_file=workflow_file_path,
    )
    logger.debug(f"PathContext: {path_context.json(indent=4)}")

    ######################################################################
    # Step 2: Update the local environment with Paths and Configuration
    # NOTE: VERY IMPORTANT TO RUN BEFORE RUNNING ANY LOCAL WORKFLOWS
    ######################################################################
    set_local_env(ws_config=ws_config, path_context=path_context)

    # Step 3: If workflow_name is provided, run single Workflow
    logger.debug(f"workflow_name: {workflow_name}")
    logger.debug(f"task_name: {task_name}")
    if workflow_name is not None and workflow_name != "":
        if workflow_name not in workflows.keys():
            logger.error(
                "Could not find '{}' in {}".format(
                    workflow_name, "[{}]".format(", ".join(workflows.keys()))
                )
            )
            return False
        workflow_to_run: Workflow = workflows[workflow_name]

        _name = workflow_to_run.name
        print_subheading(f"\nRunning Workflow: {_name}")
        # Pass down context
        workflow_to_run.run_context = run_context
        workflow_to_run.path_context = path_context
        wf_run_success = workflow_to_run.run_in_local_env(task_id=task_name)
        return wf_run_success

    # Step 4: is task name is provided, run a single task
    if task_name is not None and task_name != "":
        if tasks is None:
            logger.error("No Tasks available")
            return False

        if task_name not in tasks.keys():
            logger.error(
                "Could not find '{}' in {}".format(
                    task_name, "[{}]".format(", ".join(tasks.keys()))
                )
            )
            return False
        task_to_run: Task = tasks[task_name]

        _name = task_to_run.name
        print_subheading(f"\nRunning Task: {_name}")
        # Pass down context
        task_to_run.run_context = run_context
        task_to_run.path_context = path_context
        task_run_success = task_to_run.run_in_local_env()
        return task_run_success

    # Step 5: If workflow_name is None and task_name is None
    # means this is a file with just workflows and no data products
    # if task_name is not provided, run all workflows
    for wf_name, wf_obj in workflows.items():
        logger.debug(f"wf_name: {wf_name}")
        logger.debug(f"wf_obj: {wf_obj}")
        _name = wf_name or wf_obj.name
        print_subheading(f"\nRunning Workflow: {_name}")
        # Pass down context
        wf_obj.run_context = run_context
        wf_obj.path_context = path_context
        wf_run_success = wf_obj.run_in_local_env()
        if not wf_run_success:
            return False
    return True


def run_data_products_local(
    workflow_file: str,
    data_products: Dict[str, DataProduct],
    run_context: RunContext,
    ws_data: PhiWsData,
) -> bool:
    """
    Runs a DataProduct in the local environment

    Args:
        workflow_file: Path of the workflow file relative to the products_dir.
            This is used to build the path_context
        data_products: Dict[str, DataProduct] to run
        run_context: RunContext
        ws_data: PhiWsData

    Returns:
        run_status (bool): True is the DataProduct ran successfully
    """
    from phidata.utils.cli_console import print_run_status_table

    # Validate
    if ws_data.ws_root_path is None:
        logger.error("Workspace directory invalid")
        return False
    if ws_data.ws_config is None:
        logger.error("WorkspaceConfig invalid")
        return False

    # Step 1: Build the PathContext for the DataProducts.
    ws_config = ws_data.ws_config
    ws_root_path = ws_data.ws_root_path
    scripts_dir_path = ws_root_path.joinpath(ws_config.scripts_dir).resolve()
    storage_dir_path = ws_root_path.joinpath(ws_config.storage_dir).resolve()
    meta_dir_path = ws_root_path.joinpath(ws_config.meta_dir).resolve()
    products_dir_path = ws_root_path.joinpath(ws_config.products_dir).resolve()
    notebooks_dir_path = ws_root_path.joinpath(ws_config.notebooks_dir).resolve()
    workspace_config_dir_path = ws_root_path.joinpath(
        ws_config.workspace_config_dir
    ).resolve()
    workflow_file_path = Path(products_dir_path).joinpath(workflow_file)
    dp_path_context: PathContext = PathContext(
        scripts_dir=scripts_dir_path,
        storage_dir=storage_dir_path,
        meta_dir=meta_dir_path,
        products_dir=products_dir_path,
        notebooks_dir=notebooks_dir_path,
        workspace_config_dir=workspace_config_dir_path,
        workflow_file=workflow_file_path,
    )
    logger.debug(f"PathContext: {dp_path_context.json(indent=4)}")

    ######################################################################
    # Step 2: Update the local environment with Paths and Configuration
    # NOTE: VERY IMPORTANT TO RUN BEFORE RUNNING ANY LOCAL WORKFLOWS
    ######################################################################
    set_local_env(ws_config=ws_config, path_context=dp_path_context)

    # Step 3: Run the DataProducts
    dp_run_status: List[RunStatus] = []
    for dp_name, dp_obj in data_products.items():
        _name = dp_name or dp_obj.name
        print_subheading(f"\nRunning DataProduct: {_name}")
        # Pass down context
        dp_obj.run_context = run_context
        dp_obj.path_context = dp_path_context
        run_success = dp_obj.run_in_local_env()
        dp_run_status.append(RunStatus(_name, run_success))

    print_run_status_table("DataProduct Run Status", dp_run_status)
    for run_status in dp_run_status:
        if not run_status.success:
            return False
    return True



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, Tuple

from phidata.asset import DataAsset
from phidata.task import Task
from phidata.dq import DQCheck
from phidata.product import DataProduct
from phidata.workflow import Workflow

from phiterm.workflow.exceptions import WorkflowNotFoundException
from phiterm.utils.cli_console import print_error, print_info, print_validation_errors
from phiterm.utils.log import logger


def parse_workflow_description(
    wf_description: str, ws_root_path: Path, products_dir: str
) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """
    This function parses a workflow description string in the format:
    Dir/File:Workflow:Task and return (workflow_file, workflow_name, task_name)

    The workflow_file is the path of the file starting after the products_dir.
    The workflow_file string can then be joined to the `products_dir` in the current env
    to get the absolute workflow file path
    Examples:

    `dau` -> (dau/dau.py, None)
    `dau:save` -> (dau/dau.py, save)
    `dau:save:test` -> (dau/dau.py, save, test)
    `dau/dau2:save:test` -> (dau/dau2.py, save, test)

    Args:
        wf_description:
        ws_root_path:
        products_dir:

    Returns:
        workflow_file: string path to the file containing the workflow,
                        relative to the products dir
        workflow_name: name of the workflow (if provided)
        task_name: name of the task (if provided)
    """

    # Split the wf_description by ":"
    wf_desc_slices = wf_description.strip().split(":")
    wf_path_slice: str = wf_desc_slices[0]
    wf_py_file: str = wf_path_slice + ".py"

    workflow_file: Optional[str] = None
    workflow_name: Optional[str] = (
        wf_desc_slices[1] if len(wf_desc_slices) > 1 else None
    )
    task_name: Optional[str] = wf_desc_slices[2] if len(wf_desc_slices) > 2 else None

    # Check if products dir is valid
    products_dir_path: Path = ws_root_path.joinpath(products_dir)
    if not (products_dir_path.exists() and products_dir_path.is_dir()):
        logger.error("Invalid products dir: {}".format(products_dir_path))
        return None, None, None

    # Check if the wf_path_slice is a file or a directory
    wf_py_file_path: Path = products_dir_path.joinpath(wf_py_file)
    wf_path_slice_path: Path = products_dir_path.joinpath(wf_path_slice)

    if wf_py_file_path.exists() and wf_py_file_path.is_file():
        # the path slice is a file
        workflow_file = wf_py_file
    elif wf_path_slice_path.exists() and wf_path_slice_path.is_dir():
        # the path slice is a dir and we should get the matching file.py
        wf_file_in_dir_path = wf_path_slice_path.joinpath(wf_py_file)
        if wf_file_in_dir_path.exists() and wf_file_in_dir_path.is_file():
            workflow_file = "{}/{}".format(wf_path_slice, wf_py_file)

    if workflow_file is None:
        logger.error("Workflow file not found")
        return None, None, None

    return workflow_file, workflow_name, task_name


def get_data_products_and_workflows_from_file(
    file_path: Path,
) -> Tuple[Dict[str, DataProduct], Dict[str, Workflow]]:
    """
    Reads the DataProducts & Workflows from filepath
    Args:
        file_path:

    Returns:

    """

    import importlib.util
    from importlib.machinery import ModuleSpec
    from pydantic import ValidationError

    logger.debug(f"Reading {file_path}")

    # Read DataProduct and Workflow objects from file_path
    try:
        # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        # Create a ModuleSpec
        dp_module_spec: Optional[ModuleSpec] = importlib.util.spec_from_file_location(
            "dp_module", file_path
        )
        # Using the ModuleSpec create a module
        if dp_module_spec:
            dp_module = importlib.util.module_from_spec(dp_module_spec)
            dp_module_spec.loader.exec_module(dp_module)  # type: ignore

            # loop over all objects in module and find DataProduct and Workflow objects
            dp_dict: Dict[str, DataProduct] = OrderedDict()
            wf_dict: Dict[str, Workflow] = OrderedDict()
            for k, v in dp_module.__dict__.items():
                if isinstance(v, DataProduct):
                    logger.debug(f"Found {k} | Type: {v.__class__.__name__}")
                    dp_dict[k] = v
                if isinstance(v, Workflow):
                    logger.debug(f"Found {k} | Type: {v.__class__.__name__}")
                    wf_dict[k] = v
            logger.debug("--^^-- Loading complete")
            return dp_dict, wf_dict
    except NameError as name_err:
        print_error("Variable not defined")
        raise
    except ValidationError as validation_err:
        print_error(str(validation_err))
        # print_validation_errors(validation_err.errors())
        print_info("Please fix and try again")
        exit(0)
    except (ModuleNotFoundError, Exception) as e:
        # logger.exception(e)
        raise
    raise WorkflowNotFoundException(f"No workflow found at {file_path}")


def parse_workflow_file(
    wf_file_path: Path,
) -> Tuple[
    Dict[str, DataProduct],
    Dict[str, Workflow],
    Dict[str, Task],
    Dict[str, DataAsset],
    Dict[str, DQCheck],
]:
    """
    Reads a workflow_file and returns the
        - DataProducts
        - Workflows
        - Tasks
        - DataAssets
        - DQChecks
    Args:
        wf_file_path:

    Returns:

    """

    import importlib.util
    from importlib.machinery import ModuleSpec
    from pydantic import ValidationError

    logger.debug(f"Reading {wf_file_path}")

    # Read wf_file_path
    try:
        # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        # Create a ModuleSpec
        wf_module_spec: Optional[ModuleSpec] = importlib.util.spec_from_file_location(
            "wf_module", wf_file_path
        )
        # Using the ModuleSpec create a module
        if wf_module_spec:
            dp_module = importlib.util.module_from_spec(wf_module_spec)
            wf_module_spec.loader.exec_module(dp_module)  # type: ignore

            # loop over all objects in module and find DataProduct and Workflow objects
            dp_dict: Dict[str, DataProduct] = OrderedDict()
            wf_dict: Dict[str, Workflow] = OrderedDict()
            task_dict: Dict[str, Task] = OrderedDict()
            da_dict: Dict[str, DataAsset] = OrderedDict()
            dq_dict: Dict[str, DQCheck] = OrderedDict()
            for k, v in dp_module.__dict__.items():
                if isinstance(v, DataProduct):
                    logger.debug(f"Adding {v.__class__.__name__}: {v.name}")
                    dp_dict[v.name] = v
                if isinstance(v, Workflow):
                    logger.debug(f"Adding {v.__class__.__name__}: {v.name}")
                    wf_dict[v.name] = v
                if isinstance(v, Task):
                    logger.debug(f"Adding {v.__class__.__name__}: {v.name}")
                    task_dict[v.name] = v
                if isinstance(v, DataAsset):
                    logger.debug(f"Adding {v.__class__.__name__}: {v.name}")
                    da_dict[v.name] = v
                if isinstance(v, DQCheck):
                    logger.debug(f"Adding {v.__class__.__name__}: {v.name}")
                    dq_dict[v.name] = v

            logger.debug("--^^-- Loading complete")
            return dp_dict, wf_dict, task_dict, da_dict, dq_dict
    except NameError as name_err:
        print_error("Variable not defined")
        raise
    except ValidationError as validation_err:
        print_error(str(validation_err))
        # print_validation_errors(validation_err.errors())
        print_info("Please fix and try again")
        exit(0)
    except (ModuleNotFoundError, Exception) as e:
        # logger.exception(e)
        raise
    raise WorkflowNotFoundException(f"No workflow found at {wf_file_path}")
